#!/usr/bin/env python

##
## NOTE: ANY CHANGES HERE WILL NOT BE REFLECTED IN THE EVALUATION SETUP
##       PLEASE MAKE YOUR CHANGES IN `run.py`.
##

import os
import math
import pandas as pd

import torch
from torchvision import transforms

from torch.utils.data import Dataset

import tqdm

from skimage import io
import numpy as np
from exceptions import OutOfBudetException


class ZEWDPCBaseDataset(Dataset):
    """
    ZEW Data Purchasing Challenge Base Dataset Class
    """

    def __init__(
        self,
        images_dir,
        labels_path,
        drop_labels: bool = False,
        shuffle_seed=None,
        transform=None,
    ):
        """
        Initializes a ZEWDPCBaseDataset class.

        Args:
            images_dir      : Directory containing dataset images referenced in the annotation file
            labels_path     : Path to the CSV file containing labels for all the images in the dataset
            drop_labels     : Boolean value which determines if the labels should be dropped from the response
                              when accessing the dataset via `ZEWDPCBaseDataset.__getitem__(...)`.
            shuffle_seed    : Seed for reproducibility during the shuffle of dataset rows during initialization
            transform       : A transformation function to pre-process the images before they are served to the models.
        """
        self.images_dir = images_dir
        self.labels_path = labels_path

        self.drop_labels = drop_labels
        self.shuffle_seed = shuffle_seed  # Seed used to shuffle the rows of the dataset

        self.transform = transform

        self.labels_column_names = [
            "scratch_small",
            "scratch_large",
            "dent_small",
            "dent_large",
        ]
        self.validate_and_load_labels_file()

    def validate_and_load_labels_file(self):
        """
        Internal function to validate and load the labels file.
        """
        self.labels_df = pd.read_csv(self.labels_path)
        # Shuffle all the rows across each instantiation
        self.labels_df = self.labels_df.sample(
            frac=1, random_state=self.shuffle_seed
        ).reset_index(drop=True)

        if self.drop_labels:
            # Drop Label Columns if they are present in the labels file
            label_columns = set(self.labels_df.columns).intersection(
                self.labels_column_names
            )
            self.labels_df = self.labels_df.drop(label_columns, axis=1)

    def __len__(self):
        return len(self.labels_df)

    def _get_row(self, idx: int):
        """
        Access a row from the dataframe representing the dataset

        Args:
            idx: index of the row in the dataframe

        Returns:
            pandas.core.series.Series object representing a row in the dataset
        """
        row = self.labels_df.iloc[idx]
        return row

    def _get_label(self, idx: int):
        """
        Get the labels for the data point at a particular index.

        Args:
            idx: index of the datapoint

        Returns:
            a list representing the labels for the data point at a particular index.
            Example [0, 1, 1, 0]
        """
        idx = self._validate_idx(idx)
        row = self._get_row(idx)
        # Gather the label based on the order of the keys in the labels_column_names
        label = [row[x] for x in self.labels_column_names]
        return label

    def _get_all_labels(self):
        """
        Returns all the labels for the whole dataset.
        Useful during the evaluation for comparison with the ground_truth.
        """
        return self.labels_df[self.labels_column_names].to_numpy()

    def _get_filename(self, idx):
        row = self._get_row(idx)
        filename = row.filename.replace(".png", "").replace(".jpg", "")
        return filename

    def __getitem__(self, idx: int):
        """
        Function to access an image, label pair at a parituclar index.

        Args:
            idx: index of the datapoint

        Returns:
            A dictionary object with the following keys :
                - `image`
                    Holds a representation of the image at the referenced index.
                    It can either be the raw image, as loaded by skimage.io.imread
                    or if a `transform` function is provided, then it the transformed
                    representation of the image.
                - `label`
                    Holds a representation of the labels at the referenced index.
                    The structure of the labels are the same as returned by the `self._get_label` function.

                    NOTE: if `drop_labels` is True, then the `label` key is excluded from the dictionary.
        """
        idx = self._validate_idx(idx)

        row = self._get_row(idx)

        # Collect Image
        filename = row["filename"]
        image_path = os.path.join(self.images_dir, filename)
        image = io.imread(image_path)

        # Prepare Response Sample
        sample = {}
        sample["idx"] = idx
        sample["image"] = image
        if not self.drop_labels:
            sample["label"] = self._get_label(idx)

        # Apply Transformations
        if self.transform:
            sample["image"] = self.transform(sample["image"])

        return sample

    def set_transform(self, transform):
        """
        Function to set a transformation function for the images in the dataset.

        Args:
            transform: A transformation function which takes as input a raw image
                as loaded by `skimage.io.imread` and returns a representation of the
                image, as required by the participants.
                `torchvision.transforms` (https://pytorch.org/vision/stable/transforms.html)
                are supported.
        """
        self.transform = transform

    ########################################################################
    ########################################################################
    # Utils
    ########################################################################
    ########################################################################

    def _validate_idx(self, idx):
        if torch.is_tensor(idx):
            # `pandas.DataFrame` does not accept tensors as indices
            # so if idx is a tensor, then we convert it into an scalar idx
            idx = idx.item()
        return idx


class ZEWDPCProtectedDataset(ZEWDPCBaseDataset):
    """
    A protected Dataset access object which wraps over an `ZEWDPCBaseDataset` object
    and manages a data purchase policy for the dataset.

    This shares the same interfaces as the ZEWDPCBaseDataset class, and the key differences are :
        * The labels are not included by default in the samples when you iterate over this dataset.
        * Allows participants to iteratively purchase the labels of certain datapoints, while respecting the available
            budget.
    """

    def __init__(
        self,
        images_dir,
        labels_path,
        budget=None,
        shuffle_seed=None,
        transform=None,
    ):
        super().__init__(
            images_dir=images_dir,
            labels_path=labels_path,
            drop_labels=False,
            shuffle_seed=shuffle_seed,
            transform=transform,
        )

        self.budget = math.inf if budget == None else budget
        # Manages a set of purchased labels
        self.purchases = set()

    def purchase_label(self, idx: int):
        """
        Request to purchase the label for a sample.

        Returns the label if :
            * if there is still budget available to purchase more
        else:
            throws an exception

        NOTE: If you do not use this function to access the labels,
        and instead try to access it locally using self.__base_dataset,
        your code will not be functional in the evaluation setup.
        """
        if self.check_available_budget() > 0 or (idx in self.purchases):
            self.purchases.add(idx)
            label = self._get_label(idx)
            return label
        else:
            raise OutOfBudetException(self.budget)

    def check_available_budget(self):
        """
        Returns the available label-access budget
        """
        return self.budget - len(self.purchases)


if __name__ == "__main__":

    ###########################################################################
    ###########################################################################
    ##
    ## BaseDataset Access Examples
    ###########################################################################
    ###########################################################################
    dataset = ZEWDPCBaseDataset(
        images_dir="./data/prepared/v0.1/dataset_debug",
        labels_path="./data/prepared/v0.1/dataset_debug/labels.csv",
        drop_labels=False,
    )

    print("Labels Dictionary :", dataset.labels_column_names)
    """
    You can treat the dataset as an iterator, 
    where you can iterate over all the data samples.
    """
    for sample in tqdm.tqdm(dataset):
        """
        Each of the samples will have the following structure :

            {
                'idx': 0,
                'image': array([[[110, 128, 140],
                    [110, 128, 139],
                    [110, 128, 140],
                    ...,
                    [132, 152, 166],
                    [134, 154, 168],
                    [137, 158, 173]]]),
                'label': [0, 0, 1, 1]
            }

        where :
            `idx`   : contains the reference id for this image
            `image` : contains the image as an numpy array loaded by skimage.io.imread
            `label` : contains the associated labels for this data point.
                      The values at each of the indices in the label represent the presence or absence
                      of the following features :
                        ['scratch_small', 'scratch_large', 'dent_small', 'dent_large']

                    If `drop_labels` is passed as True during the instantiation of the class,
                    then the `labels` key is not included in the sample.
        """
        print(sample)
        break

    ###########################################################################
    ###########################################################################
    ##
    ## Protected Dataset Access Examples
    ###########################################################################
    ###########################################################################
    p_dataset = ZEWDPCProtectedDataset(
        images_dir="./data/prepared/v0.1/dataset_debug",
        labels_path="./data/prepared/v0.1/dataset_debug/labels.csv",
        budget=150,
    )

    """
    The Protected Dataset is derived from the Base Dataset,
    and is used to hold the unlabelled dataset. 
    This can be treated as an iterator as well, and can be used to 
    access the samples very much like the Base Dataset. 
    
    A Protected Dataset is always instantiated by providing a "budget", which 
    refers to the maximum number of labels that can be probed from the said 
    Protected Dataset instance. 
    
    The samples in a Protected Dataset will never contain the `label` key. 
    """
    for sample in tqdm.tqdm(p_dataset):
        idx = sample["idx"]

        # Labels, instead have to be "purchased"
        label = p_dataset.purchase_label(idx)
        print(label, p_dataset.check_available_budget())
        # When the budget for accessing the labels has been exhausted, the
        # Protected Dataset will throw an OutOfBudetException.

        if idx == 50:
            # Example of transform applied to the images in the dataset
            # midway.
            preprocess = transforms.Compose(
                [
                    transforms.ToTensor(),
                    transforms.Grayscale(num_output_channels=1),
                    transforms.Normalize(mean=[0.5], std=[0.5]),
                ]
            )
            p_dataset.set_transform(preprocess)
            print(sample.keys())
            input("Transofrm applied. Press any key....")
