#!/usr/bin/env python

import numpy as np
from tqdm.auto import tqdm

from evaluator.dataset import ZEWDPCBaseDataset, ZEWDPCProtectedDataset


class ZEWDPCBaseRun:
    """
    Template Submission Class for the ZEW Data Purchasing Challenge 2022.

    The submission template follows the following hooks :
        - pre_training_phase
        - purchase_phase
        - prediction_phase
        - save_checkpoint
        - load_checkpoint

        Please refer to the inline documentation for further details.
        You are allowed to add any other member functions, however you
        are not allowed to change the names of these hooks, else your
        submissions will not be evaluated by the automated evaluators.
    """

    def __init__(self):
        self.evaluation_state = {}

    def pre_training_phase(
        self, training_dataset: ZEWDPCBaseDataset, register_progress=lambda x: False
    ):
        """
        # Pre-training Phase
        -------------------------
        Pretrain your model on the available labelled dataset here
        Hook for the Pre-Training Phase of the Competition, where you
        have access to a training_dataset, which is an instance of the
        `ZEWDPCBaseDataset` class (see dataset.py for more details).

        You are allowed to pre-train on this data, while you prepare
        for the Purchase_Phase of the competition.

        If you train some models, you can instantiate them as `self.model`,
        as long as you implement self-contained checkpointing in the
        `self.save_checkpoint` and `self.load_checkpoint` hooks, as the
        hooks for the different phases of the competition, can be called
        in different executions of the BaseRun.
        """
        print("\n================> Pre-Training Phase\n")

        iteration = 0
        for sample in tqdm(training_dataset):
            iteration += 1

            progress = iteration / len(training_dataset)
            register_progress(progress)  # [Optional, but recommended] Mark Progress
            """
            You also have access to a `register_progress` function, to which you can
            pass a value between [0,1] to relay onto the leaderboard your self reported
            progress on the training phase to be displayed on the submission dashboard
            during the evaluation. If a value out of these bounds is provided, it will
            be clipped to this range.
            """

        print("Execution Complete of Training Phase.")

    def purchase_phase(
        self,
        unlabelled_dataset: ZEWDPCProtectedDataset,
        training_dataset: ZEWDPCBaseDataset,
        budget=1000,
        register_progress=lambda x: False,
    ):
        """
        # Purchase Phase
        -------------------------
        In this phase of the competition, you have access to
        the unlabelled_dataset (an instance of `ZEWDPCProtectedDataset`)
        and the training_dataset (an instance of `ZEWDPCBaseDataset`)
        {see datasets.py for more details}, and a purchase budget.

        You can iterate over both the datasets and access the images without restrictions.
        However, you can probe the labels of the unlabelled_dataset only until you
        run out of the label purchasing budget.
        """
        print("\n================> Purchase Phase | Budget = {}\n".format(budget))

        register_progress(0.0) #Register Progress
        
        for sample in tqdm(unlabelled_dataset):
            idx = sample["idx"]
            
            # Budgeting & Purchasing Labels
            if budget > 0:
                label = unlabelled_dataset.purchase_label(idx)
            budget -= 1

        register_progress(1.0) #Register Progress
        print("Execution Complete of Purchase Phase.")

    def prediction_phase(
        self,
        test_dataset: ZEWDPCBaseDataset,
        register_progress=lambda x: False,
    ):
        """
        # Prediction Phase
        -------------------------
        In this phase of the competition, you have access to a test set, and you
        are supposed to make predictions using your trained models.

        Returns:
            np.ndarray of shape (n, 4)
                where n is the number of samples in the test set
                and 4 refers to the 4 labels to be predicted for each sample
                for the multi-label classification problem.
        """
        print(
            "\n================> Prediction Phase : - on {} images\n".format(
                len(test_dataset)
            )
        )
        
        predictions = []
        for sample in tqdm(test_dataset):
            # Your Prediction Code here
            predictions.append(
                np.random.randint(low=0, high=2, size=4)
            )
            pass
        
        register_progress(1.0)

        predictions = np.array(predictions) # random predictions
        print("Execution Complete of Purchase Phase.")
        return predictions

    def save_checkpoint(self, checkpoint_path):
        """
        Self-contained checkpoint code to be included here,
        which can capture the state of your run (including any trained models, etc)
        at the provided path.

        This is critical to implement, as the execution of the different phases can
        happen using different instances of the BaseRun. See below for examples.
        """
        pass

    def load_checkpoint(self, checkpoint_path):
        """
        Self-contained checkpoint code to be included here,
        which can load the state of your run (including any trained models, etc)
        from a provided path (previously saved using `self.save_checkpoint`)

        This is critical to implement, as the execution of the different phases can
        happen using different instances of the BaseRun. See below for examples.
        """
        pass


if __name__ == "__main__":

    ## NOTE: The code below is only for illustration purposes.
    ####################################################################################
    ####################################################################################
    ##
    ## Dataset Initialization
    ####################################################################################
    DATASET_SHUFFLE_SEED = 1022022

    # Instantiate Training Dataset
    training_dataset = ZEWDPCBaseDataset(
        images_dir="./data/prepared/v0.1/dataset_debug",
        labels_path="./data/prepared/v0.1/dataset_debug/labels.csv",
        shuffle_seed=DATASET_SHUFFLE_SEED,
    )
    # Instantiate Unlabelled Dataset
    unlabelled_dataset = ZEWDPCProtectedDataset(
        images_dir="./data/prepared/v0.1/dataset_debug",
        labels_path="./data/prepared/v0.1/dataset_debug/labels.csv",
        budget=100,  # Configurable Parameter
        shuffle_seed=DATASET_SHUFFLE_SEED,
    )
    # Instantiate Validation Dataset
    val_dataset = ZEWDPCBaseDataset(
        images_dir="./data/prepared/v0.1/dataset_debug",
        labels_path="./data/prepared/v0.1/dataset_debug/labels.csv",
        drop_labels=True,
        shuffle_seed=DATASET_SHUFFLE_SEED,
    )
    # A second instantiation of the validation test with the labels present
    #       - helpful later, when computing the scores.
    val_dataset_gt = ZEWDPCBaseDataset(
        images_dir="./data/prepared/v0.1/dataset_debug",
        labels_path="./data/prepared/v0.1/dataset_debug/labels.csv",
        drop_labels=False,
        shuffle_seed=DATASET_SHUFFLE_SEED,
    )

    ####################################################################################
    ####################################################################################
    ##
    ## Phase 1 : Pre-Training Phase
    ####################################################################################
    run = ZEWDPCBaseRun()
    run.pre_training_phase(training_dataset)

    run.save_checkpoint("/tmp/pretrainig_phase_checkpoint.pickle")
    # NOTE:It is critical that the checkpointing works in a self-contained way
    #       As, the evaluators might choose to run the different phases separately.

    del run

    ####################################################################################
    ####################################################################################
    ##
    ## Phase 2 : Pre-Training Phase
    ####################################################################################
    run = ZEWDPCBaseRun()
    run.load_checkpoint("/tmp/pretrainig_phase_checkpoint.pickle")
    run.purchase_phase(unlabelled_dataset, training_dataset, budget=3000)

    run.save_checkpoint("/tmp/purchase_phase_checkpoint.pickle")
    del run

    ####################################################################################
    ####################################################################################
    ##
    ## Phase 3 : Prediction Phase
    ####################################################################################
    run = ZEWDPCBaseRun()
    run.load_checkpoint("/tmp/purchase_phase_checkpoint.pickle")
    predictions = run.prediction_phase(val_dataset)
    assert type(predictions) == np.ndarray
    assert predictions.shape == (len(val_dataset), 4)

    ####################################################################################
    ####################################################################################
    ##
    ## Phase 4 : Evaluation Phase
    ####################################################################################
    from evaluator.evaluation_metrics import accuracy_score, hamming_loss, exact_match_ratio

    y_true = val_dataset_gt._get_all_labels()
    y_pred = predictions

    accuracy_score = accuracy_score(y_true, y_pred)
    hamming_loss_score = hamming_loss(y_true, y_pred)
    exact_match_ratio_score = exact_match_ratio(y_true, y_pred)

    print("Accuracy Score : ", accuracy_score)
    print("Hamming Loss : ", hamming_loss_score)
    print("Exact Match Ratio : ", exact_match_ratio_score)
